package org.openxri.pipeline.stages;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Properties;
import java.util.Vector;

import org.openxri.XRISegment;
import org.openxri.config.ServerConfig;
import org.openxri.exceptions.StageException;
import org.openxri.exceptions.StoreException;
import org.openxri.factories.ServerConfigFactory;
import org.openxri.resolve.TrustType;
import org.openxri.store.Authority;
import org.openxri.store.Store;
import org.openxri.store.StoreBetterLookup;
import org.openxri.store.SubSegment;
import org.openxri.urimapper.URIMapper;
import org.openxri.xml.AuthorityResolutionService;
import org.openxri.xml.CanonicalID;
import org.openxri.xml.Service;
import org.openxri.xml.XRD;

/**
 * A stage that adds an authority resolution service endpoint to the XRD. It uses the server's URIMapper to
 * appropriately construct the URI. It is recommended to put this stage into the LOOKUP pipeline, so that 
 * community i-names can be resolved to an arbitrary depth.
 * 
 * Parameters for the stage's configuration:
 * 
 * http (optional, default: true):
 * Set to true, if a HTTP endpoint URI should be added to the service endpoint.
 * 
 * https (optional, default: true): 
 * Set to true, if a HTTPS endpoint URI should be added to the service endpoint.
 * 
 * httpport (optional, default: 80):
 * The TCP port of the HTTP endpoint URI
 * 
 * httpsport (optional, default: 443):
 * The TCP port of the HTTPS endpoint URI
 * 
 * trust (optional, default: none):
 * The trust type to use for the service endpoint's media type element.
 * 
 * check-children (optional, default: true):
 * If this is true, then the authority resolution SEP will only be added if the 
 * authority actually has child subsegments.
 * 
 * Recommended pipeline(s) for this stage: 
 * LOOKUP
 * 
 * @author =peacekeeper
 */
public class AddAuthorityResolutionSEPStage extends AbstractStage {

	public static final String PROPERTIES_KEY_HTTP = "http";
	public static final String PROPERTIES_KEY_HTTPS = "https";
	public static final String PROPERTIES_KEY_HTTPPORT = "httpport";
	public static final String PROPERTIES_KEY_HTTPSPORT = "httpsport";
	public static final String PROPERTIES_KEY_TRUST = "trust";
	public static final String PROPERTIES_KEY_CHECK_CHILDREN = "check-children";

	public static final String DEFAULT_HTTP = "true";
	public static final String DEFAULT_HTTPS = "true";
	public static final String DEFAULT_HTTPPORT = "80";
	public static final String DEFAULT_HTTPSPORT = "443";
	public static final String DEFAULT_TRUST = "none";
	public static final String DEFAULT_CHECK_CHILDREN = "true";

	private String httpURI;
	private String httpsURI;
	private TrustType trust;
	private boolean checkChildren;
	private URIMapper uriMapper;

	public AddAuthorityResolutionSEPStage(Properties properties) {

		super(properties);
	}

	public void init() throws Exception {

		ServerConfig serverConfig = ServerConfigFactory.getSingleton();

		String httpPort = this.properties.getProperty(PROPERTIES_KEY_HTTPPORT, DEFAULT_HTTPPORT);
		String httpsPort = this.properties.getProperty(PROPERTIES_KEY_HTTPSPORT, DEFAULT_HTTPSPORT);
		
		String http = this.properties.getProperty(PROPERTIES_KEY_HTTP, DEFAULT_HTTP);
		if (Boolean.parseBoolean(http)) {

			StringBuffer buffer = new StringBuffer()
			.append("http://")
			.append(serverConfig.getHostName())
			.append(":")
			.append(httpPort)
			.append(serverConfig.getServletPath());

			this.httpURI = buffer.toString();
		}

		String https = this.properties.getProperty(PROPERTIES_KEY_HTTPS, DEFAULT_HTTPS);
		if (Boolean.parseBoolean(https)) {

			StringBuffer buffer = new StringBuffer()
			.append("https://")
			.append(serverConfig.getHostName())
			.append(":")
			.append(httpsPort)
			.append(serverConfig.getServletPath());

			this.httpsURI = buffer.toString();
		}

		String trust = this.properties.getProperty(PROPERTIES_KEY_TRUST, DEFAULT_TRUST);
		this.trust = new TrustType(trust);

		String checkChildren = this.properties.getProperty(PROPERTIES_KEY_CHECK_CHILDREN, DEFAULT_CHECK_CHILDREN);
		this.checkChildren = Boolean.parseBoolean(checkChildren);

		this.uriMapper = (URIMapper) ServerConfigFactory.getSingleton().getComponentRegistry().getComponent(URIMapper.class);
	}

	public XRD execute(Store store, XRD xrd, XRISegment segment, Authority parentAuthority, String subSegmentName, Authority authority, boolean isCreate) throws StageException {

		// only add the SEP if the authority has children?

		if (store instanceof StoreBetterLookup && this.checkChildren && authority != null) {

			SubSegment[] children;

			try {

				children = ((StoreBetterLookup) store).getAuthorityChildSubSegments(authority);
				if (children.length < 1) return(xrd);
			} catch (StoreException ex) {

				throw new StageException("Cannot access store to check for child subsegments.", ex);
			}
		}

		// use the XRD's CanonicalID as the provider ID for the authority resolution service endpoint

		CanonicalID canonicalID = xrd.getCanonicalID();
		String canonicalIDString = null;

		if (canonicalID != null) {

			canonicalIDString = canonicalID.getValue();
		}

		// find out what to use as the namespace for further subsegments

		String namespace = null;

		if (canonicalIDString != null) {

			namespace = canonicalIDString;
		} else if (store instanceof StoreBetterLookup) {

			String qxris[];

			try {

				qxris = ((StoreBetterLookup) store).getAuthorityQxris(authority);
			} catch (StoreException ex) {

				throw new StageException("Cannot read subsegment XRI from store.", ex);
			}

			if (qxris.length > 0) namespace = qxris[0];
		}

		if (namespace == null) {

			throw new StageException("Cannot determine namespace for authority resolution SEP.");
		}

		// create HTTP and HTTPS <URI> elements for the authority resolution service endpoint

		Vector<URI> uris = new Vector<URI> ();

		if (this.httpURI != null) {

			StringBuffer uri = new StringBuffer(this.httpURI);
			this.uriMapper.completeURI(uri, namespace);

			try {

				uris.add(new URI(uri.toString()));
			} catch (URISyntaxException ex) {

				throw new StageException("Invalid URI: " + uri.toString());
			}
		}

		if (this.httpsURI != null) {

			StringBuffer uri = new StringBuffer(this.httpsURI);
			this.uriMapper.completeURI(uri, namespace);

			try {

				uris.add(new URI(uri.toString()));
			} catch (URISyntaxException ex) {

				throw new StageException("Invalid URI: " + uri.toString());
			}
		}

		// create authority resolution service endpoint and add it to the XRD

		Service authorityResolutionService = new AuthorityResolutionService(uris.toArray(new URI[uris.size()]), canonicalIDString, this.trust);

		xrd.addService(authorityResolutionService);

		// done

		return(xrd);
	}
}
